#pragma once

#include "Owned_string.hxx"
#include <iostream>

// Class for representing ranges of characters, usually for the
// purpose of comparing them. The idea is that we have constructors for
// a bunch of different ways that strings might be specified, and each
// gets converted to a string_view [begin, end).
//
// Enforces invariant that begin <= end; throws otherwise.
class string_view
{
public:
    // Constructs a string_view from `begin` and `end` directly.
    string_view(const char* begin, const char* end);

    // Constructs a string_view from the start and the size.
    string_view(const char*, size_t);

    // Constructs a string_view from the contents of a `String`.
    string_view(Owned_string const&);

    // Constructs a string_view from a '\0'-terminated C-style string.
    explicit string_view(const char*);

    // Constructs a string_view from a string literal using its static size.
    template <size_t N>
    string_view(const char (&s) [N]);

    /*
     * Getters
     */

    const char* begin() const;
    const char* end() const;

    size_t size() const;

    /*
     * Member variables
     */

private:
    const char* begin_;
    const char* end_;
    // INVARIANT: begin_ <= end_
};

// Overloads == for `string_view`s.
bool operator==(string_view, string_view);

// Overloads != for `string_view`s.
bool operator!=(string_view, string_view);

// Templates need to be defined in the .h file, not the .cpp file.
// (We subtract 1 from N because N will include the string literal's
// '\0' terminator.)
template <size_t N>
string_view::string_view(const char (&s) [N])
        : string_view(s, N - 1)
{ }

// Overloads stream insertion (printing)
std::ostream& operator<<(std::ostream&, string_view);

